######################################################################
# top-level Makefile for expmake of the cscope package
######################################################################

######################################################################
# If you want to run an xterm/etermx window on the remote machines (for
# interactive debugging) then change 'NO' to 'YES' for INTERACTIVE.
#
# Use the 'always' target to force running the top level rules even if
# cscope is up to date.
######################################################################
INTERACTIVE = NO
#INTERACTIVE = YES

EMULATOR = xterm
MYDISPLAY = mskeates-1.uk.lucent.com:0

######################################################################
# Source directory for all distributable sources (except this Makefile)
######################################################################
SRCDIR = "src"

######################################################################
# The compiler defines __STDC__ on the following machine types:
#
#	i486
#	sgi
#	ilinux
#	solaris
#
# __STDC__ is NOT defined on:
#
#	sparc
#	ibm
#	uts4
#	hppa
#
# If __STDC__ is defined, then generally the ANSI function prototypes
# are supported. Some functions must also have the ANSI function 
# definitions, particularly if arguments of 'char' or 'register char'
# are passed to avoid the automatic promotion to int that the old-style
# function definitions caused.
######################################################################

# TYPE and ANSICC are overridden on remote machine
TYPE = unspecified
ANSICC = cc

# "Worldwide Exptools" - tools made available on the public internet.
WORLDWIDE=-worldwide
VERSION=13.8.8

# For worldwide operation need to use gnutypes
TYPES = `$(TOOLS)/adm/bin/machines gnutypes|xargs|tr ' ' ','`

default: help

CURSES = /usr/ccs/lib/libcurses.a

debugcscope:	always
	@set -ex; \
	echo "##################################################" ; \
	TARGETS="cscope"; \
	case $(TYPE) in \
	*cygwin) \
		TARGETS="cscope.exe";; \
	esac; \
	eval make TYPE=$(TYPE) DEBUG=-g $$TARGETS; \
	mv $$TARGETS debug$$TARGETS

debugmlcscope:	always
	@set -ex; \
	echo "##################################################" ; \
	TARGETS="mlcscope"; \
	case $(TYPE) in \
	*cygwin) \
		TARGETS="mlcscope.exe";; \
	esac; \
	eval make TYPE=$(TYPE) DEBUG=-g $$TARGETS; \
	mv $$TARGETS debug$$TARGETS

cscopecpiogz:	always
	@set -ex; \
	EMULATOR="$(EMULATOR)"; export EMULATOR; \
	if [ $(INTERACTIVE) = "YES" ] ; \
	then \
		export PATH DISPLAY EDITOR XT LINES COLUMNS; \
		PATH=$$PATH:/usr/bin/X11:/usr/X11R6/bin:/opt/x11r6/bin:/usr/openwin/bin:`logdir exptools`/bin; \
		DISPLAY=$(MYDISPLAY); \
		EDITOR=emacs; \
		LINES=24; COLUMNS=80; \
		XT="-fn 6x13"; \
		$$EMULATOR $$XT -title "$(TYPE) -- EXP TESTING with $$EMULATOR"; \
		true; \
	fi; \
	echo "##################################################" ; \
	TARGETS="cscope"; \
	case $(TYPE) in \
	*cygwin) \
		TARGETS="cscope.exe mlcscope.exe";; \
	*) \
		TARGETS="cscope mlcscope";; \
	esac; \
	eval make TYPE=$(TYPE) $$TARGETS; \
	make VERSION=$(VERSION) EXEEXT=$(EXEEXT) checkversion; \
	PATH=/opt/exp/gnu/bin:$$PATH strip $$TARGETS; \
	ls $$TARGETS | cpio -ocv | gzip >cscopecpiogz; \
	if [ $(INTERACTIVE) = "YES" ] ; \
	then \
		export PATH DISPLAY EDITOR XT LINES COLUMNS; \
		PATH=$$PATH:/usr/bin/X11:/usr/X11R6/bin:/opt/x11r6/bin:/usr/openwin/bin:`logdir exptools`/bin; \
		DISPLAY=$(MYDISPLAY); \
		EDITOR=emacs; \
		LINES=24; COLUMNS=80; \
		XT="-fn 6x13"; \
		$$EMULATOR $$XT -title "$(TYPE) -- EXP TESTING with $$EMULATOR"; \
		true; \
	fi

uncpiogz:
	types=$(TYPES); \
	types=`echo $$types | tr ',' ' '`; \
	for type in $$types; do \
	if [ -f cscopecpiogz.$$type ] ; then \
	echo "uncpio $$type"; \
	gzcat cscopecpiogz.$$type | cpio -icdm; \
	if [ -f cscope ] ; then \
	mv cscope cscope.$$type; \
	rm -f bin/cscope.$$type; \
	ln cscope.$$type bin/cscope.$$type; \
	fi; \
	if [ -f cscope.exe ] ; then \
	mv cscope.exe cscope.exe.$$type; \
	rm -f bin/cscope.exe.$$type; \
	ln cscope.exe.$$type bin/cscope.exe.$$type; \
	fi; \
	if [ -f mlcscope ] ; then \
	mv mlcscope mlcscope.$$type; \
	rm -f bin/mlcscope.$$type; \
	ln mlcscope.$$type bin/mlcscope.$$type; \
	fi; \
	if [ -f mlcscope.exe ] ; then \
	mv mlcscope.exe mlcscope.exe.$$type; \
	rm -f bin/mlcscope.exe.$$type; \
	ln mlcscope.exe.$$type bin/mlcscope.exe.$$type; \
	fi; \
	fi; \
	done

mlcscope$(EXEEXT): always
	echo "##################################################" ; \
	echo "# Building cscope for target = mlcscope$(EXEEXT) #" ; \
	echo "##################################################" ; \
	cd $(SRCDIR); \
	if [ -f /opt/exp/gnu/bin/gcc ] ; then \
	CC=/opt/exp/gnu/bin/gcc; export CC; \
	fi; \
	case $(TYPE) in \
	*cygwin) \
		TARGETS="mlcscope.exe";; \
	*) \
		TARGETS="mlcscope";; \
	esac; \
	case $(TYPE) in \
	*cygwin|ilinux|*linux-gnu) \
		if [ -s stdscan.c ] ;then cp stdscan.c scanner.c; fi;; \
	esac; \
	if [ -n "$(DEBUG)" ] ; \
	then \
		TARGETS="DEBUG=$(DEBUG) $$TARGETS";  \
	fi; \
	case $(TYPE) in \
	  hppa|*hpux) make LIBS=-lPW CURSES=-lcurses HPPA=-Dhppa \
			USEVFS=U MULTILINGUAL=D $$TARGETS;; \
	  *sgi*) make LIBS=-lPW  CURSES=-lcurses USEVFS=U MULTILINGUAL=D \
			$$TARGETS;; \
	  ibm)      make LIBS=-lPW  CURSES=-lcurses SVR2=D VOID_PROTOTYPE=U \
			USEVFS=U MULTILINGUAL=D $$TARGETS;; \
	  uts4)     make CC='/usr/ccs/bin/cc -Xa' LIBS=-lPW CURSES=-lcurses \
			SVR2=U USEVFS=U MULTILINGUAL=D $$TARGETS;; \
	  i486)     make LIBS=-lgen CURSES=-lcurses USEVFS=U MULTILINGUAL=D \
			$$TARGETS;; \
	  *solaris)  make LIBS=-lgen CURSES="$(CURSES)" MULTILINGUAL=D \
			USEVFS=D $$TARGETS;; \
	  sparc|*sunos) make LIBS= CURSES=-lcurses BSD=D \
			USEVFS=U MULTILINGUAL=D LDFLAGS=-Bstatic $$TARGETS;; \
	  ilinux|*linux-gnu) make LIBS= CURSES=/usr/lib/libncurses.a MULTILINGUAL=D \
			LINUX='-I/usr/include/ncurses -Dilinux -D__linux__' \
			USEVFS=U $$TARGETS;; \
	  *cygwin)  make LIBS= CURSES=-lncurses USEVFS=U MULTILINGUAL=D \
			LEX='/bin/flex -l' FLEX=/lib/libfl.a \
			CYGWIN=-Dcygwin $$TARGETS;; \
	  *darwin)  make LIBS= CURSES=-lncurses BSD=D USEVFS=U \
			DARWIN='-I/usr/include/ncurses -Ddarwin' \
			MULTILINGUAL=D $$TARGETS;; \
	  zlinux|*ibm-linux) make LIBS= CURSES=-lncurses MULTILINGUAL=D \
			LINUX='-I/usr/include/ncurses -Dilinux -D__linux__ -Dzlinux' \
			USEVFS=U $$TARGETS;; \
	  *freebsd*)  make BSD=D LIBS=-ltermcap CURSES=-lcurses DARWIN=-Ddarwin \
			LEX='/bin/flex -l' FLEX=/lib/libfl.a \
			USEVFS=U MULTILINGUAL=D $$TARGETS;; \
	  *)        echo "Invalid target ($(TYPE)) specified";; \
	esac; \
	make checkmlcscope; \
	if [ -f mlcscope$(EXEEXT) ] ; then \
	cp mlcscope$(EXEEXT) ..; \
	fi; \
	make clean

cscope$(EXEEXT): always
	echo "##################################################" ; \
	echo "# Building cscope for target = cscope$(EXEEXT)   #" ; \
	echo "##################################################" ; \
	cd $(SRCDIR); \
	if [ -f /opt/exp/gnu/bin/gcc ] ; then \
	CC=/opt/exp/gnu/bin/gcc; export CC; \
	fi; \
	case $(TYPE) in \
	*cygwin) \
		TARGETS="cscope.exe";; \
	*) \
		TARGETS="cscope";; \
	esac; \
	case $(TYPE) in \
	*cygwin|ilinux|*linux-gnu) \
		if [ -s stdscan.c ] ;then cp stdscan.c scanner.c; fi;; \
	esac; \
	if [ -n "$(DEBUG)" ] ; \
	then \
		TARGETS="DEBUG=$(DEBUG) $$TARGETS"; \
	fi; \
	case $(TYPE) in \
	  hppa|*hpux) make LIBS=-lPW  CURSES=-lcurses HPPA=-Dhppa \
			USEVFS=U $$TARGETS;; \
	  *sgi*)    make LIBS=-lPW  CURSES=-lcurses USEVFS=U $$TARGETS;; \
	  ibm)      make LIBS=-lPW  CURSES=-lcurses SVR2=D VOID_PROTOTYPE=U \
			USEVFS=U $$TARGETS;; \
	  uts4)     make CC='/usr/ccs/bin/cc -Xa' LIBS=-lPW CURSES=-lcurses \
			SVR2=U USEVFS=U $$TARGETS;; \
	  i486)     make LIBS=-lgen CURSES=-lcurses USEVFS=U $$TARGETS;; \
	  *solaris)  make LIBS=-lgen CURSES="$(CURSES)" USEVFS=D $$TARGETS;; \
	  sparc|*sunos) make LIBS= CURSES=-lcurses BSD=D \
			USEVFS=U LDFLAGS=-Bstatic $$TARGETS;; \
	  ilinux|*linux-gnu) make LIBS=      CURSES=/usr/lib/libncurses.a \
			LINUX='-I/usr/include/ncurses -Dilinux -D__linux__' \
			USEVFS=U $$TARGETS;; \
	  *cygwin)  make LIBS=      CURSES=-lncurses USEVFS=U \
			CYGWIN=-Dcygwin $$TARGETS;; \
	  *darwin)  make LIBS= CURSES=-lncurses BSD=D USEVFS=U \
			DARWIN='-I/usr/include/ncurses -Ddarwin' \
			$$TARGETS;; \
	  zlinux|*ibm-linux) make LIBS=      CURSES=-lncurses USEVFS=U \
			LINUX='-I/usr/include/ncurses -Dilinux -D__linux__ -Dzlinux' \
			$$TARGETS;; \
	  *freebsd*)  make BSD=D LIBS=-ltermcap CURSES=-lcurses DARWIN=-Ddarwin \
			USEVFS=U $$TARGETS;; \
	  *)        echo "Invalid target ($(TYPE)) specified";; \
	esac; \
	make checkcscope; \
	if [ -f cscope$(EXEEXT) ] ; then \
	cp cscope$(EXEEXT) ..; \
	fi; \
	make clean

checkversion:
	cd $(SRCDIR); \
	make VERSION=$(VERSION) EXEEXT=$(EXEEXT) chkversion

always:

local:
	cd $(SRCDIR); make CURSES=-lcurses LIBS="-lgen" cscope

clean:
	cd $(SRCDIR); make clean

help:
	@echo "No default target"

makewww:
	cd $(SRCDIR); make stdscan.c; cd ..
	xhost +
	$(TOOLS)/adm/bin/expmake $(WORLDWIDE) version="$(VERSION)" -parallel -wait4signature executableTypes=$(TYPES)

preparewww:
	cd $(SRCDIR); make stdscan.c; cd ..
	$(TOOLS)/adm/bin/expmake $(WORLDWIDE) version="$(VERSION)" -wait4signature -prepareOnly executableTypes=$(TYPES)

providewww:
	$(TOOLS)/adm/bin/expprovide $(WORLDWIDE) version="$(VERSION)" -wait4signature executableTypes=$(TYPES) cscope

cleanwww:
	$(TOOLS)/adm/bin/expmake $(WORLDWIDE) version="$(VERSION)" -clean -parallel -skipPrepare executableTypes=$(TYPES)

savewww:
	$(TOOLS)/adm/bin/expmake $(WORLDWIDE) version="$(VERSION)" -skipPrepare executableTypes=""

makeint:
	cd $(SRCDIR); make stdscan.c; cd ..
	xhost +
	$(TOOLS)/adm/bin/expmake -parallel -wait4signature

prepareint:
	cd $(SRCDIR); make stdscan.c; cd ..
	$(TOOLS)/adm/bin/expmake -wait4signature -prepareOnly

provideint:
	$(TOOLS)/adm/bin/expprovide -wait4signature cscope

cleanint:
	$(TOOLS)/adm/bin/expmake -clean -parallel -skipPrepare

savesrc:
	$(TOOLS)/adm/bin/expmake -skipPrepare executableTypes=""

######################################################################
# Use the scanner generated with a standard lex rather than the flex
# often on other machines. This also needs to be done with DOS-based
# versions of cscope.
#
# HOWEVER, if the locally generated scanner.c is compatible, then use
# it instead of the 'standard' one on Solaris
#
# Note that under Solaris 2.8 the lex is different to 2.5 and hence
# scanner.c has extra includes that don't compile on other machines
# such as <inttypes.h>. Beware!
######################################################################
ilinux cygwin darwin:
	cd $(SRCDIR); make stdscan.c; cd ..
	xhost +
	$(TOOLS)/adm/bin/expmake -parallel -skipPrepare executableTypes=$@

hppa sgi solaris sparc isolaris:
	xhost +
	$(TOOLS)/adm/bin/expmake -parallel -skipPrepare executableTypes=$@

cleanilinux cleancygwin cleandarwin:
	$(TOOLS)/adm/bin/expmake -parallel -clean -skipPrepare executableTypes=`echo $@ | sed 's+^clean++'`

cleanhppa cleansgi cleansolaris cleansparc cleanisolaris:
	$(TOOLS)/adm/bin/expmake -parallel -clean -skipPrepare executableTypes=`echo $@ | sed 's+^clean++'`

i686-pc-linux-gnu i686-pc-cygwin powerpc-apple-darwin:
	cd $(SRCDIR); make stdscan.c; cd ..
	xhost +
	$(TOOLS)/adm/bin/expmake $(WORLDWIDE) version="$(VERSION)" -parallel -skipPrepare executableTypes=$@

hppa1.1-hp-hpux mips-sgi-irix sparc-sun-solaris sparc-sun-sunos i386-pc-solaris:
	xhost +
	$(TOOLS)/adm/bin/expmake $(WORLDWIDE) version="$(VERSION)" -parallel -skipPrepare executableTypes=$@

cleani686-pc-linux-gnu cleani686-pc-cygwin cleanpowerpc-apple-darwin cleans390-ibm-linux:
	$(TOOLS)/adm/bin/expmake $(WORLDWIDE) version="$(VERSION)" -parallel -clean -skipPrepare executableTypes=`echo $@ | sed 's+^clean++'`

cleanhppa1.1-hp-hpux cleanmips-sgi-irix cleansparc-sun-solaris cleansparc-sun-sunos cleani386-pc-solaris:
	$(TOOLS)/adm/bin/expmake $(WORLDWIDE) version="$(VERSION)" -parallel -clean -skipPrepare executableTypes=`echo $@ | sed 's+^clean++'`

