/*
 *   Copyright (C) 1988-2006 Lucent Technologies Inc. and others.
 *   All Rights Reserved.
 */

/*	cscope - interactive C symbol cross-reference
 *
 *	keyword look-up routine for the C symbol scanner
 */

#ifdef MULTILINGUAL

#include "global.h"

/* keyword text for fast testing of keywords in the scanner */
char	externtext[] = "extern";
char	typedeftext[] = "typedef";
extern	char	classtext[];

/* This keyword table is also used for keyword text compression.  Keywords
 * with an index less than the numeric value of a space are replaced with the
 * control character corresponding to the index, so they cannot be moved
 * without changing the database file version and adding compatibility code
 * for old databases.
 */
struct	keystruct ckeyword[] = {
	{ "#define",	' ',	MISC,	NULL },	/* must be table entry 0 for old databases */
	{ "#include",	' ',	MISC,	NULL },	/* must be table entry 1 */
	{ "break",	'\0',	FLOW,	NULL },	/* rarely in cross-reference */
	{ "case",	' ',	FLOW,	NULL },
	{ "char",	' ',	DECL,	NULL },
	{ "continue",	'\0',	FLOW,	NULL },	/* rarely in cross-reference */
	{ "default",	'\0',	FLOW,	NULL },	/* rarely in cross-reference */
	{ "#define",	' ',	MISC,	NULL },	/* must be table entry 7 */
	{ "double",	' ',	DECL,	NULL },
	{ "\t",		'\0',	MISC,	NULL },	/* must be table entry 9 */
	{ "\n",		'\0',	MISC,	NULL },	/* must be table entry 10 */
	{ "else",	' ',	FLOW,	NULL },
	{ "enum",	' ',	DECL,	NULL },
	{ externtext,	' ',	DECL,	NULL },
	{ "float",	' ',	DECL,	NULL },
	{ "for",	'(',	FLOW,	NULL },
	{ "goto",	' ',	FLOW,	NULL },
	{ "if",		'(',	FLOW,	NULL },
	{ "int",	' ',	DECL,	NULL },
	{ "long",	' ',	DECL,	NULL },
	{ "register",	' ',	DECL,	NULL }, /* 20 */
	{ "return",	'\0',	FLOW,	NULL },
	{ "short",	' ',	DECL,	NULL },
	{ "sizeof",	'\0',	MISC,	NULL },
	{ "static",	' ',	DECL,	NULL },
	{ "struct",	' ',	DECL,	NULL },
	{ "switch",	'(',	FLOW,	NULL },
	{ typedeftext,	' ',	DECL,	NULL },
	{ "union",	' ',	DECL,	NULL },
	{ "unsigned",	' ',	DECL,	NULL },
	{ "void",	' ',	DECL,	NULL }, /* 30 */
	{ "while",	'(',	FLOW,	NULL },
	
	/* these keywords are not compressed */
	{ "asm",	'(',	MISC,	NULL },
	{ "auto",	' ',	DECL,	NULL },
	{ "catch",	'(',	FLOW,	NULL },
	{ classtext,	' ',	DECL,	NULL },
	{ "const",	' ',	DECL,	NULL },
	{ "delete",	' ',	MISC,	NULL },
	{ "do",		' ',	FLOW,	NULL },
	{ "fortran",	' ',	DECL,	NULL },
	{ "friend",	' ',	DECL,	NULL },
	{ "inline",	' ',	DECL,	NULL },
	{ "new",	' ',	MISC,	NULL },
	{ "operator",	'\0',	DECL,	NULL },
	{ "overload",	' ',	DECL,	NULL },
	{ "private",	'\0',	DECL,	NULL },
	{ "protected",	'\0',	DECL,	NULL },
	{ "public",	'\0',	DECL,	NULL },
	{ "signed",	' ',	DECL,	NULL },
	{ "template",	' ',	DECL,	NULL },
	{ "this",	' ',	MISC,	NULL },
	{ "throw",	' ',	FLOW,	NULL },
	{ "try",	' ',	FLOW,	NULL },
	{ "virtual",	' ',	DECL,	NULL },
	{ "volatile",	' ',	DECL,	NULL },
};
#define KEYWORDS	(sizeof(ckeyword) / sizeof(struct keystruct))

#define HASHMOD	(KEYWORDS * 2 + 1)

static	struct	keystruct *hashtab[HASHMOD]; /* pointer table */

#if defined( __STDC__ ) || defined( __BORLANDC__ )
void	cinitsymtab(void);
struct	keystruct *clookup(char *ident);
#endif

/* put the keywords into the symbol table */

void
cinitsymtab()
{
	register int	i, j;
	register struct	keystruct *p;
	
	for (i = 1; i < KEYWORDS; ++i) {
		p = &ckeyword[i];
		j = hash(p->text) % HASHMOD;
		p->next = hashtab[j];
		hashtab[j] = p;
	}
}

/* see if this identifier is a keyword */

struct	keystruct *
clookup(ident)
register char	*ident;
{
	register struct	keystruct *p;
	int	c;
	
	/* look up the identifier in the keyword table */
	for (p = hashtab[hash(ident) % HASHMOD]; p != NULL; p = p->next) {
		if (strequal(ident, p->text)) {
			if (compress == YES && (c = p - ckeyword) < ' ') {
				ident[0] = c;	/* compress the keyword */
			}
			return(p);
		}
	}
	/* this is an identifier */
	return(NULL);
}

#endif
